from collections import defaultdict
from typing import Union

import numpy as np

from rllib._base import _BaseAgent


class QLearningAgent(_BaseAgent):
    """
    Q-Learning Agent.
    """

    def __init__(
        self,
        alpha: float,
        epsilon: float,
        discount: float,
        n_actions: int,
    ):
        """
        Q-Learning Agent Initialization.

        Args:
            alpha (float): learning rate.
            epsilon (float): exploration probability.
            discount (float): discount rate (aka gamma).
            n_actions (int): number of possible actions.
        """

        self.alpha = alpha
        self.epsilon = epsilon
        self.discount = discount
        self.n_actions = n_actions

        self._qvalues = defaultdict(lambda: defaultdict(lambda: 0.0))  # type: ignore

    def _get_qvalue(
        self,
        state: int,
        action: int,
    ) -> float:
        """
        Get Q(state, action)
        """

        return self._qvalues[state][action]

    def _set_qvalue(
        self,
        state: int,
        action: int,
        value: float,
    ) -> None:
        """
        Set Q(state, action) = value
        """

        self._qvalues[state][action] = value

    def _get_value(
        self,
        state: int,
    ) -> float:
        """
        Compute agent's estimate of V(s) using current q-values
        V(s) = max_over_action[Q(state, action)] over possible actions.
        """

        possible_actions = range(self.n_actions)

        if len(possible_actions) == 0:
            return 0.0

        value = max([self._get_qvalue(state, action) for action in possible_actions])
        return value

    def get_action(
        self,
        state: int,
    ) -> Union[None, int]:
        """
        Compute action in a state, including exploration.
        """

        possible_actions = range(self.n_actions)

        if len(possible_actions) == 0:
            return None

        action = np.random.choice(
            a=[
                self.get_best_action(state),
                np.random.choice(possible_actions),
            ],
            p=[
                1 - self.epsilon,
                self.epsilon,
            ],
        )

        return int(action)

    def get_best_action(
        self,
        state: int,
    ) -> Union[None, int]:
        """
        Compute the best action to take in a state.
        """

        possible_actions = range(self.n_actions)

        if len(possible_actions) == 0:
            return None

        best_action = np.argmax(
            [self._get_qvalue(state, action) for action in possible_actions]
        )
        return int(best_action)

    def update(
        self,
        state: int,
        action: int,
        reward: float,
        next_state: int,
    ) -> None:
        """
        Q-Value update:
            Q(s,a) := (1 - alpha) * Q(s,a) + alpha * (r + gamma * V(s'))
        """

        q_value = (1 - self.alpha) * self._get_qvalue(state, action) + self.alpha * (
            reward + self.discount * self._get_value(next_state)
        )

        self._set_qvalue(state, action, q_value)


class EVSarsaAgent(QLearningAgent):
    """
    Expected Value SARSA Agent.
    """

    def _get_value(
        self,
        state: int,
    ) -> float:
        """
        Compute agent's estimate of V(s) using current q-values
        V(s) = sum _{over a_i} {pi(a_i | s) * Q(s, a_i)}
        """

        possible_actions = range(self.n_actions)

        if len(possible_actions) == 0:
            return 0.0

        q_values = [self._get_qvalue(state, action) for action in possible_actions]
        max_q_value_idx = np.argmax(q_values)

        value = 0.0

        for i, q_value in enumerate(q_values):
            if i == max_q_value_idx:
                value += (
                    (1 - self.epsilon) + self.epsilon / len(possible_actions)
                ) * q_value
            else:
                value += (self.epsilon / len(possible_actions)) * q_value

        return value



from abc import ABC, abstractmethod


class _BaseAgent(ABC):
    """
    Base RL Agent.
    """

    @abstractmethod
    def get_action(self, state):
        ...

    @abstractmethod
    def get_best_action(self, state):
        ...

    @abstractmethod
    def update(self, state, action, reward, next_state):
        ...



import random

import numpy as np
from gym import Env


def set_global_seed(
    seed: int,
    env: Env,
):
    """
    Set global seed for reproducibility.

    Args:
        seed (int): seed.
        env (Env): gym environment.
    """

    random.seed(seed)
    np.random.seed(seed)
    env.seed(seed)



from setuptools import setup

from rllib import __version__

with open("README.md", mode="r", encoding="utf-8") as fp:
    long_description = fp.read()


setup(
    name="pytorch-rllib",
    version=__version__,
    description="Reinforcement Learning Library.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Dani El-Ayyass",
    author_email="dayyass@yandex.ru",
    license_files=["LICENSE"],
    url="https://github.com/dayyass/rllib",
    packages=["rllib"],
    install_requires=[
        "gym==0.24.1",
        "numpy==1.21.6",
        "tqdm==4.64.0",
    ],
)



from typing import List

from gym import Env
from tqdm import trange

from rllib._base import _BaseAgent


class Trainer:
    """
    Class to train agent in environment.
    """

    def __init__(
        self,
        env: Env,
    ):
        """
        Trainer initialization with gym environment.

        Args:
            env (Env): gym environment.
        """

        self.env = env

    def train(
        self,
        agent: _BaseAgent,
        n_sessions: int,
        epsilon_decay: float = 0.99,
        t_max: int = 10**4,
    ) -> List[float]:
        """
        Train loop over sessions.

        Args:
            agent (_BaseAgent): RL agent.
            n_sessions (int): number of sessions to train.
            epsilon_decay (float, optional): epsilon decay. Defaults to 0.99.
            t_max (int, optional): max number of one session actions. Defaults to 10**4.

        Returns:
            List[float]: rewards over sessions.
        """

        rewards = []
        for _ in trange(n_sessions, desc="loop over sessions"):
            r = self._play_session(
                agent=agent,
                t_max=t_max,
            )
            rewards.append(r)

            agent.epsilon *= epsilon_decay  # type: ignore

        return rewards

    def _play_session(
        self,
        agent: _BaseAgent,
        t_max: int,
    ) -> float:
        """
        One session cycle.

        Args:
            agent (_BaseAgent): RL agent.
            t_max (int): max number of one session actions.

        Returns:
            float: session reward.
        """

        total_reward = 0.0
        s = self.env.reset()

        for _ in range(t_max):
            a = agent.get_action(s)
            next_s, r, done, _ = self.env.step(a)

            agent.update(s, a, r, next_s)

            s = next_s
            total_reward += r
            if done:
                break

        return total_reward
